# -*- coding: utf-8 -*-
"""

    Name: Strategy name
    Version: 1.0.0

"""
import math
import time
from datetime import datetime

from enum import Enum

from arquants import Strategy, Order


class Req(Enum):
    NEW = 0
    REPLACE = 1
    CANCEL = 2


class StrategyType(Enum):
    MM = 0


class MMStrategyState(Enum):
    WAITING_FOR_SIGNAL = 0
    SENDING_ORDERS = 1
    WAITING_TRADE = 2
    HOLD = 3


class MMStrategy(Strategy):

    def __init__(self, min_spread_ticks=5, order_size=100, debug=True):


        super(MMStrategy, self).__init__()
        #self.data2
        strategies = []
        states = []
        md = {
            StrategyType.MM.value: [[]]
        }
        self.initialize(strategies, states, md, debug)

        # Parameters

    def generate_index_mm(self):
        """
        return a dictionary by index with a list of instruments associated with the index.
        every index represent a unique set of instruments for the strategy
        :return: dict of list of list
        """
        index = 0
        data = 0
        sets = 1
        subsets = 1
        index_list = dict()
        while True:
            if hasattr(self, 'data' + str(data)):
                for i in range(subsets):
                    index_list[index] = ['data' + str(data + i)]
                    index += 1
            else:
                break
            data += sets
        return index_list

    def generate_signal_mm(self, index, data):
        self.log_md(StrategyType.MM, index)
        if self.trades_to_process[StrategyType.MM.value][index] > 0:
            self.trades_to_process[StrategyType.MM.value][index] -= 1
            self.logs_if("Strategy: %s - Index: %s - MD trade generados por la estrategia" % (StrategyType.MM.name, index))
        elif self.strategy_state[StrategyType.MM.value][index] is MMStrategyState.WAITING_FOR_SIGNAL.value:
            if self.get_data_value(data[0], "bid_px") and self.get_data_value(data[0], "offer_px"):
                bid_px = data[0].bid_px[0] + data[0].get_tick()
                offer_px = data[0].offer_px[0] - data[0].get_tick()
                if offer_px - bid_px >= (self.min_spread_ticks * data[0].get_tick()):
                    bid_qty, offer_qty = self.get_size()
                    self.signals[StrategyType.MM.value][index] = []
                    signal = dict()
                    signal['type'] = 0
                    signal['px'] = bid_px
                    signal['qty'] = bid_qty
                    signal['side'] = Order.Buy
                    self.signals[StrategyType.MM.value][index].append(signal)
                    signal = dict()
                    signal['type'] = 0
                    signal['px'] = offer_px
                    signal['qty'] = offer_qty
                    signal['side'] = Order.Sell
                    self.signals[StrategyType.MM.value][index].append(signal)
                else:
                    self.logs_if("Strategy: %s - Index: %s - Min tick no se cumple" % (StrategyType.MM.name, index))
            else:
                self.logs_if("Strategy: %s - Index: %s - No hay md en el book" % (StrategyType.MM.name, index))
        elif self.strategy_state[StrategyType.MM.value][index] is MMStrategyState.WAITING_TRADE.value:
            offer_px = data[0].offer_px[0]
            bid_px = data[0].bid_px[0]
            offer_signal = None
            bid_signal = None
            for ord in self.my_orders[StrategyType.MM.value][index].values():
                if ord.ordtype == Order.Buy:
                    if not self.my_compare_float(bid_px, ord.price):
                        bid_px = bid_px + data[0].get_tick()
                        bid_signal = dict()
                        bid_signal['type'] = 1
                        bid_signal['replace'] = ord.m_orderId
                        bid_signal['px'] = bid_px
                        bid_signal['qty'] = ord.size
                if ord.ordtype == Order.Sell:
                    if not self.my_compare_float(offer_px, ord.price):
                        offer_px = offer_px - data[0].get_tick()
                        offer_signal = dict()
                        offer_signal['type'] = 1
                        offer_signal['replace'] = ord.m_orderId
                        offer_signal['px'] = offer_px
                        offer_signal['qty'] = ord.size
            if bid_signal or offer_signal:
                if len(self.my_orders[StrategyType.MM.value][index]) == 2:
                    if offer_px - bid_px >= (self.min_spread_ticks * data[0].get_tick()):
                        self.signals[StrategyType.MM.value][index] = []
                        if bid_signal:
                            self.signals[StrategyType.MM.value][index].append(bid_signal)
                        if offer_signal:
                            self.signals[StrategyType.MM.value][index].append(offer_signal)
                    else:
                        self.logs_if("Strategy: %s - Index: %s - Min tick no se cumple, cancelamos." %
                                     (StrategyType.MM.name, index))
                        self.signals[StrategyType.MM.value][index] = []
                        for ord in self.my_orders[StrategyType.MM.value][index].values():
                            cancel_signal = dict()
                            cancel_signal['type'] = 2
                            cancel_signal['cancel'] = ord.m_orderId
                            self.signals[StrategyType.MM.value][index].append(cancel_signal)
                else:
                    self.signals[StrategyType.MM.value][index] = []
                    if offer_signal:
                        self.signals[StrategyType.MM.value][index].append(offer_signal)
                    if bid_signal:
                        self.signals[StrategyType.MM.value][index].append(bid_signal)
            else:
                self.logs_if("Strategy: %s - Index: %s - No podemos mejorar." % (StrategyType.MM.name, index))

    def process_signal_mm(self, index, signals):
        self.change_state(index, StrategyType.MM.value, MMStrategyState.SENDING_ORDERS.value)
        for signal in signals:
            if signal['type'] == 0:
                self.new_order(StrategyType.MM, index, signal['side'], 0, signal['px'], signal['qty'])
            elif signal['type'] == 1:
                ord_to_replace = self.my_orders[StrategyType.MM.value][index][signal['replace']]
                self.replace_order(StrategyType.MM, index, ord_to_replace, signal['px'], signal['qty'])
            elif signal['type'] == 2:
                ord_to_cancel = self.my_orders[StrategyType.MM.value][index][signal['cancel']]
                self.cancel_order(StrategyType.MM, index, ord_to_cancel)

    def notify_order_mm(self, index, order):
        if self.strategy_state[StrategyType.MM.value][index] in (MMStrategyState.SENDING_ORDERS.value,
                                                                 MMStrategyState.WAITING_TRADE.value):
            if order.status in (Order.Completed, Order.Partial):
                self.acum[order.ordtype] += abs(list(order.executed.exbits)[-1].size)
            # Order.Completed, Order.Partial, Order.Accepted, Order.Rejected, Order.Cancelled, Order.Canceled
            if not self.my_orders[StrategyType.MM.value][index]:
                self.change_state(index, StrategyType.MM.value, MMStrategyState.WAITING_FOR_SIGNAL.value)
            elif not self.ack[StrategyType.MM.value][index][0] and \
                    not self.ack[StrategyType.MM.value][index][1] and \
                    not self.ack[StrategyType.MM.value][index][2]:
                self.change_state(index, StrategyType.MM.value, MMStrategyState.WAITING_TRADE.value)

    def get_size(self):
        diff = self.acum[Order.Buy] - self.acum[Order.Sell]
        if self.my_compare_float(diff, 0.0):
            return self.order_size, self.order_size
        elif diff > 0:
            return self.order_size - diff, self.order_size
        elif diff < 0:
            return self.order_size, self.order_size + diff

    def initialize(self, strategies, states, md, debug=True):

        # Parameters
        self.strategies = strategies  # enum with sub-strategies
        self.states = states  # list order by sub-strategy id with enum of sub-strategies states
        self.md_info = md  # Dictionary by sub-strategy with a list of list of market data to check
        self.debug = debug  # used for logging
        self.log_string = "" # used to store logs during event

        self.idx_to_process = dict()  # Dictionary by strategy with a set of index to process when md arrives.
        self.idx_data = dict()  # Dictionary by index with a list of instrument associated to the index
        self.idx_by_inst = dict() # Dictionary by instrument's data with list of index associated to the instrument
        self.symbol_data = dict() # Dictionary by strategy and index with list of data associated
        self.signals = dict()  # Dictionary by strategy and index with signals generated by the strategy
        self.strategy_state = dict()  # Dictionary by strategy and index with the state of the strategy
        self.last_md = dict()  # Dictionary by strategy and index with last market data received
        self.trades_to_process = dict()  # Dictionary by strategy and index of number of trades to process
        self.my_orders = dict()  # Dictionary by strategy and index of orders by order_id
        self.index_of_orders = dict()  # Dictionary by strategy and order_id with index of the order
        # List of dictionaries by strategy, index and order_id
        # where we store information about new request sent to the market (0=new, 1=replace, 2=cancel)
        self.ack = [dict(), dict(), dict()]

        self.log_string = ""    # log string

        self.init_data()
        self.log_index_info()

    def next(self):
        self.generate_signals()

        # process signals
        for strategy in self.strategies:
            for index, signals in self.signals[strategy.value].items():
                getattr(self, 'process_signal_%s' % strategy.name.lower())(index, signals)

        # clear processed signals
        for strategy in self.strategies:
            self.signals[strategy.value].clear()

        self.flush_logs()

    def new_order(self, strategy, index, side, symbol_id, px, qty):
        if side == Order.Buy:
            order = self.buy(data=self.symbol_data[strategy.value][index][symbol_id],
                             price=px, size=qty, exectype=Order.Limit)
        elif side == Order.Sell:
            order = self.sell(data=self.symbol_data[strategy.value][index][symbol_id],
                              price=px, size=qty, exectype=Order.Limit)

        self.index_of_orders[strategy.value][order.m_orderId] = index
        self.my_orders[strategy.value][index][order.m_orderId] = order
        self.ack[strategy.value][index][Req.NEW.value][order.m_orderId] = order

    def replace_order(self, strategy, index, order_to_replace, new_px, new_qty):
        replaced_order = self.replace(size=new_qty, price=new_px, order=order_to_replace)

        self.index_of_orders[strategy.value][replaced_order.m_orderId] = index
        self.my_orders[strategy.value][index][replaced_order.m_orderId] = replaced_order
        self.ack[strategy.value][index][Req.REPLACE.value][replaced_order.m_orderId] = order_to_replace

    def cancel_order(self, strategy, index, order_to_cancel):
        self.cancel(order_to_cancel)
        self.ack[strategy.value][index][Req.CANCEL.value][order_to_cancel.m_orderId] = order_to_cancel

    def notify_order(self, order):
        # check strategy that send the order
        for strategy in self.strategies:
            if order.m_orderId in self.index_of_orders[strategy.value].keys():
                if order.status not in (Order.Created, Order.Submitted):
                    index = self.index_of_orders[strategy.value][order.m_orderId]
                    self.log_order_info(strategy.name, index, order)
                    if order.m_orderId in self.ack[strategy.value][index][Req.NEW.value].keys():
                        del self.ack[strategy.value][index][Req.NEW.value][order.m_orderId]
                    elif order.m_orderId in self.ack[strategy.value][index][Req.REPLACE.value].keys():
                        if order.status is not Order.Rejected:
                            replaced = self.ack[strategy.value][index][Req.REPLACE.value][order.m_orderId]
                            del self.my_orders[strategy.value][index][replaced.m_orderId]
                            del self.index_of_orders[strategy.value][replaced.m_orderId]
                        del self.ack[strategy.value][index][Req.REPLACE.value][order.m_orderId]
                    elif order.m_orderId in self.ack[strategy.value][index][Req.CANCEL.value].keys():
                        del self.ack[strategy.value][index][Req.CANCEL.value][order.m_orderId]
                    if order.status in (Order.Completed, Order.Rejected, Order.Canceled):
                        del self.my_orders[strategy.value][index][order.m_orderId]
                        del self.index_of_orders[strategy.value][order.m_orderId]
                    getattr(self, 'notify_order_%s' % strategy.name.lower())(index, order)
                break
            else:
                self.log_order_not_expected(strategy, None, order)
        self.flush_logs()

    def generate_signals(self):
        self.update_market_data()
        for strategy in self.strategies:
            for idx in self.idx_to_process[strategy.value]:
                # for every sub-strategy we call generate signals for every index in which market data changed
                getattr(self, 'generate_signal_%s' % strategy.name.lower())(idx, self.symbol_data[strategy.value][idx])
            self.idx_to_process[strategy.value].clear()

    def update_market_data(self):
        """
        Update market data of every instrument and identify index of every strategy that need to be process.
        """
        for instr in self.last_md.keys():
            data = getattr(self, instr)
            for md in self.last_md[instr].keys():
                self.last_md[instr][md][0] = self.last_md[instr][md][1]
                self.last_md[instr][md][1] = self.get_data_value(data, md)
                if self.last_md[instr][md][0] != self.last_md[instr][md][1]:
                    for strategy in self.strategies:
                        if instr in self.idx_by_inst[strategy.value].keys() and \
                                md in self.idx_by_inst[strategy.value][instr].keys():
                            for idx in self.idx_by_inst[strategy.value][instr][md]:
                                self.idx_to_process[strategy.value].add(idx)

    def on_pause(self):
        self.logs("Se apreto pausa, cancelando órdenes")
        self.cancel_all()
        self.flush_logs()

    def on_error(self):
        self.logs("Ocurrio un error, cancelando órdenes")
        self.cancel_all()
        self.flush_logs()

    def cancel_all(self):
        # Esperamos 1seg para evitar limite de logueo
        time.sleep(1)
        for strategy in self.strategies:
            for index in self.my_orders[strategy.value].keys():
                for order in self.my_orders[strategy.value][index].values():
                    self.logs("Cancelling order: strategy: %s - index: %s - id: %s" %
                              (strategy.name, index, order.m_orderId))
                    self.cancel(order)
            # volvemos a setaer estados iniciales
            for index in self.strategy_state[strategy.value].keys():
                self.change_state(index, strategy.value, 0)

    def init_data(self):
        for strategy in self.strategies:
            i = 0
            while True:
                if hasattr(self, 'data' + str(i)):
                    self.last_md['data' + str(i)] = dict()
                else:
                    break
                i += 1
            self.idx_by_inst[strategy.value] = dict()
            self.idx_to_process[strategy.value] = set()
            self.signals[strategy.value] = dict()
            self.strategy_state[strategy.value] = dict()
            self.trades_to_process[strategy.value] = dict()
            self.my_orders[strategy.value] = dict()
            self.index_of_orders[strategy.value] = dict()
            self.ack[strategy.value] = dict()
            self.idx_data[strategy.value] = dict()
            self.symbol_data[strategy.value] = dict()
        for strategy in self.strategies:
            result = getattr(self, 'generate_index_%s' % strategy.name.lower())()
            for idx in result.keys():
                self.symbol_data[strategy.value][idx] = []
                self.idx_data[strategy.value][idx] = []
                i = 0
                for inst in result[idx]:
                    self.symbol_data[strategy.value][idx].append(getattr(self, inst))
                    self.idx_data[strategy.value][idx].append(inst)
                    if inst not in self.idx_by_inst[strategy.value].keys():
                        self.idx_by_inst[strategy.value][inst] = dict()
                        #self.idx_by_inst[strategy.value][inst] = set()
                    for md in self.md_info[strategy.value][i]:
                        self.last_md[inst][md] = [None, None]
                        if md not in self.idx_by_inst[strategy.value][inst].keys():
                            self.idx_by_inst[strategy.value][inst][md] = set()
                        self.idx_by_inst[strategy.value][inst][md].add(idx)
                    i += 1
                self.ack[strategy.value][idx] = [None, None, None]
                self.ack[strategy.value][idx][Req.NEW.value] = dict()
                self.ack[strategy.value][idx][Req.REPLACE.value] = dict()
                self.ack[strategy.value][idx][Req.CANCEL.value] = dict()
                self.trades_to_process[strategy.value][idx] = 0
                self.strategy_state[strategy.value][idx] = 0
                self.my_orders[strategy.value][idx] = dict()

    ####################
    # HELPER FUNCTIONS #
    ####################

    def change_state(self, index, strategy_id, new_state_id):
        if self.debug:
            self.logs("Strategy: %s - Index: %s - Changing State from '%s' to '%s'" %
                  (strategy_id, index, self.strategy_state[strategy_id][index], new_state_id))
        self.strategy_state[strategy_id][index] = new_state_id

    @staticmethod
    def my_compare_float(float1, float2, prec=6):
        """ Comparamos 2 float con un minimos de precision.
        True si son igual - else false
        """
        if float1 and float2:
            return abs(float1 - float2) <= (1 / 10 ** prec)
        else:
            return float1 == float2

    @staticmethod
    def get_data_value(data, level):
        value = None
        if hasattr(data, level):
            value = getattr(data, level)[0]
            if not value or math.isnan(value):
                value = None
            else:
                value = round(value, 6)
        return value

    def get_strategy_state(self, strategy_id, index):
        return self.states[strategy_id](self.strategy_state[strategy_id][index])

    def log_md(self, strategy, index):
        if self.debug:
            line = "Strategy: %s - Index: %s - State: %s. Market Data: " % \
                   (strategy.name, index, self.get_strategy_state(strategy.value, index).name)
            i = 0
            for data_name in self.idx_data[strategy.value][index]:
                data = getattr(self, data_name)
                line = line + "%s: " % data.tradecontract
                for md in self.md_info[strategy.value][i]:
                    line = line + " %s@%s " % (md, self.last_md[data_name][md][1])
                i += 1
            self.logs(line)

    def log_order_not_expected(self, strategy, index, order):
        """
        Logging unexpected order.
        """
        state = None if not index else self.strategy_state[strategy.value][index].name
        self.logs("Strategy: %s - Index: %s - State: %s. Unexpected order: id: %s - status: %s" %
                  (strategy.name, index, state, order.m_orderId, order.status))

    def log_order_info(self, strategy_name, index, order):
        if order.status is Order.Created:
            self.logs_if("Strategy: %s - Index: %s - order Created: %s" % (strategy_name, index, order.m_orderId))
        elif order.status is Order.Submitted:
            self.logs_if("Strategy: %s - Index: %s  order Submitted: %s" % (strategy_name, index, order.m_orderId))
        elif order.status is Order.Accepted:
            self.logs_if("Strategy: %s - Index: %s  order Accepted: %s" % (strategy_name, index, order.m_orderId))
        elif order.status is Order.Partial:
            self.logs_if("Strategy: %s - Index: %s  order Partial: %s" % (strategy_name, index, order.m_orderId))
        elif order.status is Order.Completed:
            self.logs_if("Strategy: %s - Index: %s - order Completed: %s" % (strategy_name, index, order.m_orderId))
        elif order.status is Order.Canceled:
            self.logs_if("Strategy: %s - Index: %s  order Canceled: %s" % (strategy_name, index, order.m_orderId))
        elif order.status is Order.Expired:
            self.logs_if("Strategy: %s - Index: %s  order Expired: %s" % (strategy_name, index, order.m_orderId))
        elif order.status is Order.Margin:
            self.logs_if("Strategy: %s - Index: %s  order Margin: %s" % (strategy_name, index, order.m_orderId))
        elif order.status is Order.Rejected:
            self.logs_if("Strategy: %s - Index: %s  order Rejected: %s" % (strategy_name, index, order.m_orderId))
        else:
            self.logs_if("Strategy: %s - Index: %s  order with not knowing status: %s" % (strategy_name, index,
                                                                                          order.m_orderId))

    def log_index_info(self):
        for strategy in self.strategies:
            for idx in self.idx_data[strategy.value].keys():
                i = 0
                line = "Strategy: %s - Index: %s - \n" % (strategy.name, idx)
                for data in self.symbol_data[strategy.value][idx]:
                    line = line + "%s: %s\n" % (i, data.tradecontract)
                    i += 1
                self.logs_if(line)

    def logs_if(self, msg):
        if self.debug:
            self.log_string += str(datetime.now()) + " {" + msg + "}|"

    def flush_logs(self):
        self.logs(self.log_string)
        self.log_string = ""

    def get_best_bids(self, index_for_ci, data_ci):
        """
        Devuelve una lista con el 1er y 2do mejor precio BID,
        sin tener en cuenta nuestra orden.
        None, si no hay otro precio.
        """
        level = 2
        bids_d = dict()
        data = self.get_data_value(data_ci, 'bid_px')
        if data:
            bids_d[data] = data_ci.bid_qty[0]
            data = self.get_data_value(data_ci, 'bid_px_' + str(level))
        while data:
            data = round(data, 6)
            bids_d[data] = self.get_data_value(data_ci, 'bid_qty_' + str(level))
            level += 1
            data = self.get_data_value(data_ci, 'bid_px_' + str(level))
        if self.orders_ci[0][index_for_ci]:
            for my_order in self.orders_ci[0][index_for_ci].values():
                px = round(my_order.price, 6)
                if bids_d[px]:
                    if bids_d[px] >= my_order.executed.remsize:
                        bids_d[px] -= my_order.executed.remsize
        bids = [None, None]
        i = 0
        for k, v in bids_d.items():
            if bids_d[k] > 0:
                bids[i] = bids_d[k]
                i += 1
            if i == 2:
                break
        print(bids)
        return bids


# -*- coding: utf-8 -*-
"""

    Name: Generic Strategy for Fz
    Version: 0.0.1

"""

import math
import time
from datetime import datetime

from arquants import Strategy, Order
from arquants.enums import Req, Statistics, Side


class FzStrategy(Strategy):

    def __init__(self, strategies, states, md, debug=True):

        super(FzStrategy, self).__init__()

        # Parameters
        self.strategies = strategies  # enum with sub-strategies
        self.states = states  # list order by sub-strategy id with enum of sub-strategies states
        self.md_info = md  # Dictionary by sub-strategy with a list of list of market data to check
        self.debug = debug  # used for logging

        self.rejected_counter = dict() # Dictionary by strategy and index with number of rejected orders.
        self.rejected_limit = dict()    # Dictionary by strategy and index with number of rejected orders.
        self.rejected = dict()  # Dictionary by order_id with counter of rejected request for the order.
        self.idx_to_process = dict()  # Dictionary by strategy with a set of index to process when md arrives.
        self.idx_data = dict()  # Dictionary by index with a list of instrument associated to the index
        self.idx_by_inst = dict()  # Dictionary by instrument's data with list of index associated to the instrument
        self.symbol_data = dict()  # Dictionary by strategy and index with list of data associated
        self.signals = dict()  # Dictionary by strategy and index with signals generated by the strategy
        self.strategy_state = dict()  # Dictionary by strategy and index with the state of the strategy
        self.last_md = dict()  # Dictionary by strategy and index with last market data received
        self.trades_to_process = dict()  # Dictionary by strategy and index of number of trades to process
        self.my_orders = dict()  # Dictionary by strategy and index of orders by order_id
        self.index_of_orders = dict()  # Dictionary by strategy and order_id with index of the order
        self.amount = dict()  # Dictionary by strategy, index and instrument of CumQty and AvgPx

        # List of dictionaries by strategy, index and order_id
        # where we store information about new request sent to the market (0=new, 1=replace, 2=cancel)
        self.ack = [dict(), dict(), dict()]

        self.log_string = ""    # log string

        self.init_data()
        self.log_index_info()

    def next(self):
        self.generate_signals()

        # process signals
        for strategy in self.strategies:
            for index, signals in self.signals[strategy.value].items():
                getattr(self, 'process_signal_%s' % strategy.name.lower())(index, signals)

        # clear processed signals
        for strategy in self.strategies:
            self.signals[strategy.value].clear()

    def new_order(self, strategy, index, side, symbol_id, px, qty):
        if side == Order.Buy:
            order = self.buy(data=self.symbol_data[strategy.value][index][symbol_id],
                             price=px, size=qty, exectype=Order.Limit)
        elif side == Order.Sell:
            order = self.sell(data=self.symbol_data[strategy.value][index][symbol_id],
                              price=px, size=qty, exectype=Order.Limit)

        self.index_of_orders[strategy.value][order.m_orderId] = index
        self.my_orders[strategy.value][index][order.m_orderId] = order
        self.ack[strategy.value][index][Req.NEW.value][order.m_orderId] = order

    def replace_order(self, strategy, index, order_to_replace, new_px, new_qty):
        replaced_order = self.replace(size=new_qty, price=new_px, order=order_to_replace)

        self.index_of_orders[strategy.value][replaced_order.m_orderId] = index
        self.my_orders[strategy.value][index][replaced_order.m_orderId] = replaced_order
        self.ack[strategy.value][index][Req.REPLACE.value][replaced_order.m_orderId] = order_to_replace

    def cancel_order(self, strategy, index, order_to_cancel):
        self.cancel(order_to_cancel)
        self.ack[strategy.value][index][Req.CANCEL.value][order_to_cancel.m_orderId] = order_to_cancel

    def notify_order(self, order):
        # check strategy that send the order
        for strategy in self.strategies:
            if order.m_orderId in self.index_of_orders[strategy.value].keys():
                if order.status not in (Order.Created, Order.Submitted):
                    index = self.index_of_orders[strategy.value][order.m_orderId]
                    self.log_order_info(strategy.name, index, order)
                    if order.status in (Order.Completed, Order.Canceled):
                        self.update_amount(strategy.value, index, order)
                    if order.status is Order.Rejected:
                        self.rejected_counter[strategy.value][index] += 1
                    if order.m_orderId in self.ack[strategy.value][index][Req.NEW.value].keys():
                        del self.ack[strategy.value][index][Req.NEW.value][order.m_orderId]
                    elif order.m_orderId in self.ack[strategy.value][index][Req.REPLACE.value].keys():
                        replaced = self.ack[strategy.value][index][Req.REPLACE.value][order.m_orderId]
                        if order.status is Order.Rejected:
                            if replaced.m_orderId in self.rejected.keys():
                                self.cancel_order(strategy, index, replaced)
                            else:
                                self.rejected[replaced.m_orderId] = 1 # one order only could be rejected ones
                        del self.ack[strategy.value][index][Req.REPLACE.value][order.m_orderId]
                    elif order.m_orderId in self.ack[strategy.value][index][Req.CANCEL.value].keys():
                        ord_to_cancel = self.ack[strategy.value][index][Req.CANCEL.value][order.m_orderId]
                        if order.status is Order.Rejected:
                            if ord_to_cancel.m_orderId in self.rejected.keys():
                                del self.my_orders[strategy.value][index][ord_to_cancel.m_orderId]
                                del self.index_of_orders[strategy.value][index][ord_to_cancel.m_orderId]
                                del self.rejected[ord_to_cancel.m_orderId]
                                self.logs("ERROR: Orden no existe mas en el mercado. ID: %s" % ord_to_cancel.m_orderId)
                            else:
                                self.rejected[ord_to_cancel.m_orderId] = 1
                        else:
                            if ord_to_cancel.m_orderId in self.rejected.keys():
                                self.logs("ERROR: Orden cancelada por reemplazos rejectados. ID: %s" % ord_to_cancel.m_orderId)
                        del self.ack[strategy.value][index][Req.CANCEL.value][order.m_orderId]
                    if order.status in (Order.Completed, Order.Rejected, Order.Canceled):
                        del self.my_orders[strategy.value][index][order.m_orderId]
                        del self.index_of_orders[strategy.value][order.m_orderId]
                    getattr(self, 'notify_order_%s' % strategy.name.lower())(index, order)
                break
            else:
                self.log_order_not_expected(strategy, None, order)

    def generate_signals(self):
        self.update_market_data()
        for strategy in self.strategies:
            for idx in self.idx_to_process[strategy.value]:
                # for every sub-strategy we call generate signals for every index in which market data changed
                getattr(self, 'generate_signal_%s' % strategy.name.lower())(idx, self.symbol_data[strategy.value][idx])
            self.idx_to_process[strategy.value].clear()

    def update_market_data(self):
        """
        Update market data of every instrument and identify index of every strategy that need to be process.
        """
        for instr in self.last_md.keys():
            data = getattr(self, instr)
            for md in self.last_md[instr].keys():
                self.last_md[instr][md][0] = self.last_md[instr][md][1]
                self.last_md[instr][md][1] = self.get_data_value(data, md)
                if self.last_md[instr][md][0] != self.last_md[instr][md][1]:
                    for strategy in self.strategies:
                        if instr in self.idx_by_inst[strategy.value].keys() and \
                                md in self.idx_by_inst[strategy.value][instr].keys():
                            for idx in self.idx_by_inst[strategy.value][instr][md]:
                                self.idx_to_process[strategy.value].add(idx)

    def on_pause(self):
        self.logs("Se apreto pausa, cancelando órdenes")
        self.cancel_all()

    def on_error(self):
        self.logs("Ocurrio un error, cancelando órdenes")
        self.cancel_all()

    def cancel_all(self):
        # Esperamos 1seg para evitar limite de logueo
        time.sleep(1)
        for strategy in self.strategies:
            for index in self.my_orders[strategy.value].keys():
                for order in self.my_orders[strategy.value][index].values():
                    self.logs("Cancelling order: strategy: %s - index: %s - id: %s" %
                              (strategy.name, index, order.m_orderId))
                    self.cancel(order)
            # volvemos a setaer estados iniciales
            for index in self.strategy_state[strategy.value].keys():
                self.change_state(index, strategy.value, 0)

    def init_data(self):
        for strategy in self.strategies:
            i = 0
            while True:
                if hasattr(self, 'data' + str(i)):
                    self.last_md['data' + str(i)] = dict()
                else:
                    break
                i += 1
            self.amount[strategy.value] = dict()
            self.idx_by_inst[strategy.value] = dict()
            self.idx_to_process[strategy.value] = set()
            self.signals[strategy.value] = dict()
            self.strategy_state[strategy.value] = dict()
            self.trades_to_process[strategy.value] = dict()
            self.my_orders[strategy.value] = dict()
            self.index_of_orders[strategy.value] = dict()
            self.ack[strategy.value] = dict()
            self.idx_data[strategy.value] = dict()
            self.symbol_data[strategy.value] = dict()
        for strategy in self.strategies:
            result = getattr(self, 'generate_index_%s' % strategy.name.lower())()
            for idx in result.keys():
                i = 0
                self.amount[strategy.value][idx] = dict()
                self.amount[strategy.value][Statistics.Amount.value] = 0
                self.symbol_data[strategy.value][idx] = []
                self.idx_data[strategy.value][idx] = []
                for inst in result[idx]:
                    self.amount[strategy.value][idx][inst] = dict()
                    self.amount[strategy.value][idx][Statistics.Amount.value] = 0
                    self.amount[strategy.value][idx][inst][Side.BUY.value] = dict()
                    self.amount[strategy.value][idx][inst][Side.SELL.value] = dict()
                    self.amount[strategy.value][idx][inst][Statistics.Amount.value] = 0
                    self.amount[strategy.value][idx][inst][Side.BUY.value][Statistics.CumQty.value] = 0
                    self.amount[strategy.value][idx][inst][Side.BUY.value][Statistics.AvgPx.value] = 0
                    self.amount[strategy.value][idx][inst][Side.BUY.value][Statistics.Amount.value] = 0
                    self.amount[strategy.value][idx][inst][Side.SELL.value][Statistics.CumQty.value] = 0
                    self.amount[strategy.value][idx][inst][Side.SELL.value][Statistics.AvgPx.value] = 0
                    self.amount[strategy.value][idx][inst][Side.SELL.value][Statistics.Amount.value] = 0
                    self.symbol_data[strategy.value][idx].append(getattr(self, inst))
                    self.idx_data[strategy.value][idx].append(inst)
                    if inst not in self.idx_by_inst[strategy.value].keys():
                        self.idx_by_inst[strategy.value][inst] = dict()
                        # self.idx_by_inst[strategy.value][inst] = set()
                    for md in self.md_info[strategy.value][i]:
                        self.last_md[inst][md] = [None, None]
                        if md not in self.idx_by_inst[strategy.value][inst].keys():
                            self.idx_by_inst[strategy.value][inst][md] = set()
                        self.idx_by_inst[strategy.value][inst][md].add(idx)
                    i += 1
                self.ack[strategy.value][idx] = [None, None, None]
                self.ack[strategy.value][idx][Req.NEW.value] = dict()
                self.ack[strategy.value][idx][Req.REPLACE.value] = dict()
                self.ack[strategy.value][idx][Req.CANCEL.value] = dict()
                self.trades_to_process[strategy.value][idx] = 0
                self.strategy_state[strategy.value][idx] = 0
                self.my_orders[strategy.value][idx] = dict()

    ####################
    # HELPER FUNCTIONS #
    ####################

    def change_state(self, index, strategy_id, new_state_id):
        if self.debug:
            self.logs("Strategy: %s - Index: %s - Changing State from '%s' to '%s'" %
                      (strategy_id, index, self.strategy_state[strategy_id][index], new_state_id))
        self.strategy_state[strategy_id][index] = new_state_id

    @staticmethod
    def my_compare_float(float1, float2, prec=6):
        """ Comparamos 2 float con un minimos de precision.
        True si son igual - else false
        """
        if float1 and float2:
            return abs(float1 - float2) <= (1 / 10 ** prec)
        else:
            return float1 == float2

    @staticmethod
    def get_data_value(data, level):
        value = None
        if hasattr(data, level):
            value = getattr(data, level)[0]
            if not value or math.isnan(value):
                value = None
            else:
                value = round(value, 6)
        return value

    def get_strategy_state(self, strategy_id, index):
        return self.states[strategy_id](self.strategy_state[strategy_id][index])

    def log_md(self, strategy, index):
        if self.debug:
            line = "Strategy: %s - Index: %s - State: %s. Market Data: " % \
                   (strategy.name, index, self.get_strategy_state(strategy.value, index).name)
            i = 0
            for data_name in self.idx_data[strategy.value][index]:
                data = getattr(self, data_name)
                line = line + "%s: " % data.tradecontract
                for md in self.md_info[strategy.value][i]:
                    line = line + " %s@%s " % (md, self.last_md[data_name][md][1])
                i += 1
            self.logs(line)

    def log_order_not_expected(self, strategy, index, order):
        """
        Logging unexpected order.
        """
        state = None if not index else self.strategy_state[strategy.value][index].name
        self.logs("Strategy: %s - Index: %s - State: %s. Unexpected order: id: %s - status: %s" %
                  (strategy.name, index, state, order.m_orderId, order.status))

    def log_order_info(self, strategy_name, index, order):
        if order.status is Order.Created:
            self.logs_if("Strategy: %s - Index: %s - order Created: %s" % (strategy_name, index, order.m_orderId))
        elif order.status is Order.Submitted:
            self.logs_if("Strategy: %s - Index: %s  order Submitted: %s" % (strategy_name, index, order.m_orderId))
        elif order.status is Order.Accepted:
            self.logs_if("Strategy: %s - Index: %s  order Accepted: %s" % (strategy_name, index, order.m_orderId))
        elif order.status is Order.Partial:
            self.logs_if("Strategy: %s - Index: %s  order Partial: %s" % (strategy_name, index, order.m_orderId))
        elif order.status is Order.Completed:
            self.logs_if("Strategy: %s - Index: %s - order Completed: %s" % (strategy_name, index, order.m_orderId))
        elif order.status is Order.Canceled:
            self.logs_if("Strategy: %s - Index: %s  order Canceled: %s" % (strategy_name, index, order.m_orderId))
        elif order.status is Order.Expired:
            self.logs_if("Strategy: %s - Index: %s  order Expired: %s" % (strategy_name, index, order.m_orderId))
        elif order.status is Order.Margin:
            self.logs_if("Strategy: %s - Index: %s  order Margin: %s" % (strategy_name, index, order.m_orderId))
        elif order.status is Order.Rejected:
            self.logs_if("Strategy: %s - Index: %s  order Rejected: %s" % (strategy_name, index, order.m_orderId))
        else:
            self.logs_if("Strategy: %s - Index: %s  order with not knowing status: %s" % (strategy_name, index,
                                                                                          order.m_orderId))

    def log_index_info(self):
        for strategy in self.strategies:
            for idx in self.idx_data[strategy.value].keys():
                i = 0
                line = "Strategy: %s - Index: %s - \n" % (strategy.name, idx)
                for data in self.symbol_data[strategy.value][idx]:
                    line = line + "%s: %s - minsize: %s\n" % (i, data.tradecontract, data.contractsize)
                    i += 1
                self.logs_if(line)

    def update_amount(self, strategy_id, idx, order):
        last = list(order.executed.exbits)[-1]
        avgpx = ((self.amount[strategy_id][idx][order.data][order.size][Statistics.AvgPx.value] * self.amount[strategy_id][idx][order.data][order.size][Statistics.CumQty.value]) + (abs(last.size) * abs(last.price))) / (self.amount[strategy_id][idx][order.data][order.size][Statistics.CumQty.value] + abs(last.size))
        self.amount[strategy_id][idx][order.data][order.size][Statistics.CumQty.value] += abs(last.size)
        self.amount[strategy_id][idx][order.data][order.size][Statistics.AvgPx.value] = avgpx

        self.amount[strategy_id][idx][order.data][Statistics.Amount.value] +=  (last.size * last.price / order.data)
        self.amount[strategy_id][Statistics.Amount.value]
        self.amount[strategy_id][idx][Statistics.Amount.value]


        self.amount[strategy_id][idx][order.data][order.size][Statistics.Amount.value]
        pass

    def logs_if(self, msg):
        if self.debug:
            self.log_string += str(datetime.now()) + " {" + msg + "}|"

    def flush_logs(self):
        self.logs(self.log_string)
        self.log_string = ""




from .mock import Strategy
from .mock import Order
from .mock import MarketData
from .mock import Settle
from .mock import OrderLimit
from .mock import Instruments
from .mock import Trade
from .sheets import *


# -*- coding: utf-8 -*-
"""
    arqLogger: Python websocker logger for Arquants strategies.

    This modules expose functions and enumerations of the library.
"""
from .arquant_logging import ArquantLogger
from .websocket_server import start_websocket

__version__ = "0.0.1b11"


from datetime import datetime
import queue
import math
import threading
import json
import time

from apscheduler.schedulers.background import BackgroundScheduler
import websocket

from .enums import Order, Side
from .logs import StrategyLog, InitializeEventLog, EREventLog, MDEventLog
from .logs import PauseEventLog, ErrorEventLog, InternalEventLog
from .logs import NewOrderResponseLog, CancelOrderResponseLog, ReplaceOrderResponseLog


class ArquantLogger():

    def __init__(self, url, strategy, strategy_id, client_id, param_list):
        # Connection related variables
        self.url = url
        # Initializing a queue
        self.queue = queue.Queue()
        self.ws_connection = None
        self.ws_thread = None
        self.schedTask = None
        self.connected = False
        self.strategy = strategy
        self.strategy_id = strategy_id
        self.client_id = client_id
        self.param_list = param_list
        self.execution_id = round(datetime.now().timestamp() * 100)
        self.msg_id = 0
        self.event_id = 0
        self.retry = 0
        self.retry_time = 5
        self.task_id = "logging_task"
        self._log_initialize_event()

    def connect(self):
        self.strategy.logs("Starting websocket connection...")
        self.ws_connection = websocket.WebSocketApp(self.url,
                                                    on_message=self._on_message,
                                                    on_error=self._on_error,
                                                    on_close=self._on_close,
                                                    on_open=self._on_open)

        # Create a thread and target it to the run_forever function, then start it.
        self.ws_thread = threading.Thread(target=self.ws_connection.run_forever,
                                          kwargs={"ping_interval": 270})
        self.ws_thread.start()

    def _start_logging(self):
        self.schedTask = BackgroundScheduler()
        self.schedTask.add_job(self._check_logs, 'interval', seconds=1, id=self.task_id)
        self.schedTask.start()
        self.strategy.logs("Start Scheduling Logs")

    def _check_logs(self):
        if self.ws_connection.sock and self.ws_connection.sock.connected:
            while not self.queue.empty():
                msg = self.queue.get()
                self._send_log(msg)
        else:
            self.schedTask.remove_job(self.task_id)
            self.schedTask = None

    def _send_log(self, log):
        self.msg_id += 1
        log.id = self.msg_id
        self.ws_connection.send(json.dumps(log.__dict__, cls=Encoder))

    def _on_message(self, message):
        self.strategy.logs("message arrived: %s" % message)

    def _on_error(self, exception):
        self.strategy.logs("error during websocket connection: %s. "
                           "waiting %s sec until next retrying connecting." % (exception, self.retry_time))
        time.sleep(self.retry_time)
        if not self.ws_connection.sock.connected:
            self.strategy.logs("retrying connecting nro: %s." % self.retry)
            self.retry += 1
            self.connect()

    def _on_close(self):
        self.connected = False

    def _on_open(self):
        self.connected = True
        self._start_logging()

    def close_connection(self):
        if self.ws_connection.sock and self.ws_connection.sock.connected:
            self.ws_connection.close()

    def _append_log(self, log):
        self.queue.put(log)

    # Logging methods
    def _log_initialize_event(self, additional=None):
        self.event_id += 1
        self._append_log(InitializeEventLog(-1,
                                            self.strategy_id,
                                            self.execution_id,
                                            self.client_id,
                                            self.event_id,
                                            self.param_list,
                                            additional))

    def log_er_event(self, order, additional=None):
        self.event_id += 1
        try:
            last = []
            if order.executed.exbits:
                for l in list(order.executed.exbits):
                    last.append({"px": l.price, "qty": l.size})
            self._append_log(EREventLog(-1,
                                        self.strategy_id,
                                        self.execution_id,
                                        self.client_id,
                                        self.event_id,
                                        order.m_orderId,
                                        self.get_order_status(order.status),
                                        last,
                                        order.executed.remsize,
                                        additional))
        except Exception as e:
            self.strategy.logs("Could not generate log in log_er_event. exception: %s" % e)

    def log_md_event(self, data_list=None, entry_list=None, additional=None):
        self.event_id += 1
        try:
            md_map = dict()
            idx = 0
            for data in data_list:
                md_map[data.tradecontract] = dict()
                for entry in entry_list[idx]:
                    md_map[data.tradecontract][entry] = self.get_data_entry(data, entry)
                idx += 1
            self._append_log(MDEventLog(-1,
                                        self.strategy_id,
                                        self.execution_id,
                                        self.client_id,
                                        self.event_id,
                                        md_map,
                                        additional))
        except Exception as e:
            self.strategy.logs("Could not generate log in log_md_event. exception: %s" % e)

    def log_pause_event(self, description=None, additional=None):
        self.event_id += 1
        try:
            self._append_log(PauseEventLog(-1,
                                           self.strategy_id,
                                           self.execution_id,
                                           self.client_id,
                                           self.event_id,
                                           description,
                                           additional))
            time.sleep(1)
            self.close_connection()
        except Exception as e:
            self.strategy.logs("Could not generate log in log_pause_event. exception: %s" % e)

    def log_error_event(self, description=None, additional=None):
        self.event_id += 1
        try:
            self._append_log(ErrorEventLog(-1,
                                           self.strategy_id,
                                           self.execution_id,
                                           self.client_id,
                                           self.event_id,
                                           description,
                                           additional))
            self.close_connection()
        except Exception as e:
            self.strategy.logs("Could not generate log in log_error_event. exception: %s" % e)

    def log_internal_event(self, description=None, additional=None):
        self.event_id += 1
        try:
            self._append_log(InternalEventLog(-1,
                                              self.strategy_id,
                                              self.execution_id,
                                              self.client_id,
                                              self.event_id,
                                              description,
                                              additional))
        except Exception as e:
            self.strategy.logs("Could not generate log in log_internal_event. exception: %s" % e)

    def log_new_order_response(self, order, additional=None):
        try:
            self._append_log(NewOrderResponseLog(-1,
                                                 self.strategy_id,
                                                 self.execution_id,
                                                 self.client_id,
                                                 self.event_id,
                                                 order.m_orderId,
                                                 order.m_action,
                                                 order.price,
                                                 order.size,
                                                 order.data.tradecontract,
                                                 additional))
        except Exception as e:
            self.strategy.logs("Could not generate log in log_new_order_response. exception: %s" % e)

    def log_cancel_order_response(self, order, additional=None):
        try:
            self._append_log(CancelOrderResponseLog(-1,
                                                    self.strategy_id,
                                                    self.execution_id,
                                                    self.client_id,
                                                    self.event_id,
                                                    order.m_orderId,
                                                    additional))
        except Exception as e:
            self.strategy.logs("Could not generate log in log_cancel_order_response. exception: %s" % e)

    def log_replace_order_response(self, order, additional=None):
        try:
            self._append_log(ReplaceOrderResponseLog(-1,
                                                     self.strategy_id,
                                                     self.execution_id,
                                                     self.client_id,
                                                     self.event_id,
                                                     order.m_orderId,
                                                     order.m_action,
                                                     order.price,
                                                     order.size,
                                                     order.data.tradecontract,
                                                     additional))
        except Exception as e:
            self.strategy.logs("Could not generate log in log_replace_order_response. exception: %s" % e)

    def log_strategy(self, data):
        try:
            self._append_log(StrategyLog(-1,
                                         self.strategy_id,
                                         self.execution_id,
                                         self.client_id,
                                         self.event_id,
                                         data))
        except Exception as e:
            self.strategy.logs("Could not generate log in log_strategy. exception: %s" % e)

    def new_limit_order(self, data, side, px, qty, additional=None):
        if side is Side.Buy:
            order = self.strategy.buy(data=data, price=px, size=qty, exectype=2)
        elif side is Side.Sell:
            order = self.strategy.sell(data=data, price=px, size=qty, exectype=2)
        else:
            raise Exception("Invalid side: %s" % side)
        self.log_new_order_response(order, additional)
        return order

    def replace_order(self, order, new_px, new_qty, additional=None):
        replaced_order = self.strategy.replace(size=new_qty, price=new_px, order=order)
        self.log_replace_order_response(order, additional)
        return replaced_order

    def cancel_order(self, order, additional=None):
        self.strategy.cancel(order)
        self.log_cancel_order_response(order, additional)

    @staticmethod
    def get_data_entry(data, entry):
        value = None
        if hasattr(data, entry):
            value = getattr(data, entry)
            if not isinstance(value, float):
                value = value[0]
            if not value or math.isnan(value):
                value = None
            else:
                value = round(value, 6)
        return value

    @staticmethod
    def get_order_status(status):
        if status is Order.Created:
            return "Created"
        elif status is Order.Submitted:
            return "Submitted"
        elif status is Order.Accepted:
            return "Accepted"
        elif status is Order.Partial:
            return "Partial"
        elif status is Order.Completed:
            return "Completed"
        elif status is Order.Canceled:
            return "Cancelled"
        elif status is Order.Rejected:
            return "Rejected"
        elif status is Order.Expired:
            return "Expired"
        else:
            return "Unknown status: %s" % status


class Encoder(json.JSONEncoder):
    def default(self, obj):
        attr = ""
        try:
            attr = json.JSONEncoder.default(self, obj)
        except Exception:
            print("could not serialize: %s" % obj)
            attr = "--error-serializing--"
        finally:
            return attr



class SheetItem:
    def __init__(self, name, price_size, min_size, tick_px, settlement, bid_px=0, bid_qty=0, offer_px=0,
                 offer_qty=0):
        self.price_size = float(price_size)
        self.tradecontract = name
        self.tick_px = tick_px
        self.contractsize = min_size
        self.settlement = settlement.value
        self.offer_px = [0]
        self.bid_px = [0]
        self.offer_qty = [0]
        self.bid_qty = [0]
        self.offer_px[0] = float(offer_px)
        self.offer_qty[0] = float(offer_qty)
        self.bid_px[0] = float(bid_px)
        self.bid_qty[0] = float(bid_qty)



import math


def my_compare_float(float1, float2, prec=6):
    """ Comparamos 2 float con un minimos de precision.
    True si son igual - else false
    """
    if float1 and float2:
        return abs(float1 - float2) <= (1 / 10 ** prec)
    else:
        return float1 == float2


def get_data_value(data, level):
    value = None
    if hasattr(data, level):
        value = getattr(data, level)[0]
        if math.isnan(value):
            value = None
        else:
            value = round(value, 6)
    return value


def is_better(new_px, px, side):
    if side == Order.Buy:
        return new_px > px
    elif side == Order.Sell:
        return new_px < px
    return None


def round_qty(qty, min_size):
    final = qty // min_size * min_size
    extra = qty - final
    if extra > (min_size / 2):
        return final + min_size
    else:
        return final



import logging
from src.arqLogger import start_websocket
from datetime import datetime
from logging.handlers import RotatingFileHandler

logger = logging.getLogger("Rotating Log")
logger.setLevel(logging.INFO)

# add a rotating handler
handler = RotatingFileHandler("arquants.log", maxBytes=1000000, backupCount=5)
logger.addHandler(handler)


def on_new_log(log):
    logger.info("%s - %s" % (datetime.now(), log))


start_websocket(8001, on_new_log)



import time
from src.arqLogger import ArquantLogger
from arquants import Strategy, OrderLimit, Settle, MarketData, Trade

data1 = MarketData("GGAL - 48hs", 1, 1, 0.05, Settle.S_48)
data2 = MarketData("AY24 - CI", 100, 1, 0.5, Settle.S_CI)
order = OrderLimit(100.2, 50, 0, data1)
order.executed.exbits.append(Trade(order.price, 50))


def start_client(param1=10, param2=20, param3="valor3"):
   params_list = locals()
   stgy = Strategy()
   client = ArquantLogger("ws://localhost:8001/", stgy, "Testing", "bono usd vs ars", params_list)
   client.connect()

   # MSG 1
   client.log_strategy({"description": "Testing logging"})
   time.sleep(2)
   # MSG 2
   client.log_new_order_response(order, {"additional1": "strategy tasa 1", "additional2": 0})
   time.sleep(1)
   client.log_cancel_order_response(order, {"additional1": "strategy tasa 1", "additional2": 0})
   client.log_replace_order_response(order, {"additional1": "strategy tasa 1", "additional2": 0})

   time.sleep(3)
   # MSG 3
   client.log_md_event([data1, data2], [["bid_px", "bid_qty"], ["offer_px", "offer_qty"]], {"additional1": "strategy tasa 1", "additional2": 0})
   client.log_er_event(order, {"additional1": "strategy tasa 1", "additional2": 0})
   time.sleep(2)
   client.log_internal_event("description of internal", {"additional1": "strategy tasa 1", "additional2": 0})
   client.log_pause_event("description of pause", {"additional1": "strategy tasa 1", "additional2": 0})
   client.log_error_event("description of error", {"additional1": "strategy tasa 1", "additional2": 0})

   time.sleep(1000)
   client.close_connection()


if __name__ == '__main__':
   start_client()
   print("finalizing")


from datetime import datetime
from .enums import EventTypes, LogTypes, ResponseTypes
from .exceptions import DataException


class Log:
    def __init__(self, id, strategy_id, execution_id, client_id, log_type, additional):
        self.id = id
        self.strategy_id = strategy_id
        self.execution_id = execution_id
        self.client_id = client_id
        self.type = log_type
        self.timestamp = datetime.now().timestamp()
        if additional:
            if isinstance(additional, dict):
                for k, v in additional.items():
                    self.__setattr__(k, v)
            else:
                raise DataException("invalid additional type: %s. Parameter 'additional' must be an instance of dict." % type(additional))


class EventLog(Log):
    def __init__(self, id, strategy_id, execution_id, client_id, event_id, event_type, additional):
        super().__init__(id, strategy_id, execution_id, client_id, LogTypes.Event.value, additional)
        self.event_id = event_id
        self.event_type = event_type


class ResponseLog(Log):
    def __init__(self, id, strategy_id, execution_id, client_id, event_id, response_type, additional):
        super().__init__(id, strategy_id, execution_id, client_id, LogTypes.Response.value, additional)
        self.event_id = event_id
        self.response_type = response_type


class StrategyLog(Log):
    def __init__(self, id, strategy_id, execution_id, client_id, event_id, data):
        super().__init__(id, strategy_id, execution_id, client_id, LogTypes.Own.value, None)
        self.event_id = event_id
        for k, v in data.items():
            self.__setattr__(k, v)


class NewOrderResponseLog(ResponseLog):
    def __init__(self, id, strategy_id, execution_id, client_id, event_id, order_id, side, px, qty, instrument, additional):
        super().__init__(id, strategy_id, execution_id, client_id, event_id, ResponseTypes.NewOrder.value, additional)
        self.order_id = order_id
        self.side = side
        self.px = px
        self.qty = qty
        self.instrument = instrument


class CancelOrderResponseLog(ResponseLog):
    def __init__(self, id, strategy_id, execution_id, client_id, event_id, order_id, additional):
        super().__init__(id, strategy_id, execution_id, client_id, event_id, ResponseTypes.CancelOrder.value, additional)
        self.order_id = order_id


class ReplaceOrderResponseLog(ResponseLog):
    def __init__(self, id, strategy_id, execution_id, client_id, event_id, order_id, side, new_px, new_qty, instrument, additional):
        super().__init__(id, strategy_id, execution_id, client_id, event_id, ResponseTypes.ReplaceOrder.value, additional)
        self.order_id = order_id
        self.side = side
        self.new_px = new_px
        self.new_qty = new_qty
        self.instrument = instrument


class MDEventLog(EventLog):
    def __init__(self, id, strategy_id, execution_id, client_id, event_id, md_received, additional):
        super().__init__(id, strategy_id, execution_id, client_id, event_id, EventTypes.MarketData.value, additional)
        self.md_received = md_received


class EREventLog(EventLog):
    def __init__(self, id, strategy_id, execution_id, client_id, event_id, order_id, state, last, rem_size, additional):
        super().__init__(id, strategy_id, execution_id, client_id, event_id, EventTypes.ExecReport.value, additional)
        self.order_id = order_id
        self.state = state
        self.last = last
        self.rem_size = rem_size


class PauseEventLog(EventLog):
    def __init__(self, id, strategy_id, execution_id, client_id, event_id, description, additional):
        super().__init__(id, strategy_id, execution_id, client_id, event_id, EventTypes.Pause.value, additional)
        self.description = description


class ErrorEventLog(EventLog):
    def __init__(self, id, strategy_id, execution_id, client_id, event_id, description, additional):
        super().__init__(id, strategy_id, execution_id, client_id, event_id, EventTypes.Error.value, additional)
        self.description = description


class InternalEventLog(EventLog):
    def __init__(self, id, strategy_id, execution_id, client_id, event_id, description, additional):
        super().__init__(id, strategy_id, execution_id, client_id, event_id, EventTypes.Internal.value, additional)
        self.description = description


class InitializeEventLog(EventLog):
    def __init__(self, id, strategy_id, execution_id, client_id, event_id, param_list, additional):
        super().__init__(id, strategy_id, execution_id, client_id, event_id, EventTypes.Initialize.value, additional)

        if param_list:
            if isinstance(param_list, dict):
                for k, v in param_list.items():
                    self.__setattr__(k, v)
            else:
                raise DataException(
                    "invalid param_list type: %s. Parameter 'param_list' must be an instance of dict." % type(
                        param_list))




from datetime import datetime

from autobahn.twisted.websocket import WebSocketServerProtocol, WebSocketServerFactory
from twisted.internet import reactor


def log(msg):
    print("{}: {}".format(str(datetime.now()), msg))


class LoggerServer(WebSocketServerProtocol):
    """Dummy websocket protocol"""
    def __init__(self, on_message):
        super(LoggerServer, self).__init__()
        self.on_message = on_message

    def onConnect(self, request):
        log("Client connecting: {0}".format(request.peer))

    def onOpen(self):
        log("WebSocket connection open.")

    def onMessage(self, payload, isBinary):
        if isBinary:
            log("Binary message received: {0} bytes".format(len(payload)))
        else:
            message = payload.decode('utf8')
            self.on_message(message)

    def onClose(self, wasClean, code, reason):
        log("WebSocket connection closed: {}".format(reason))

    def connectionLost(self, reason):
        WebSocketServerProtocol.connectionLost(self, reason)


class WsProtocolFactory(WebSocketServerFactory):

    def __init__(self, on_message):
        super(WsProtocolFactory, self).__init__()
        self.on_message = on_message

    def buildProtocol(self, *args, **kwargs):
        protocol = LoggerServer(self.on_message)
        protocol.factory = self
        return protocol


def start_websocket(port, on_message):
    factory = WsProtocolFactory(on_message)
    # factory.setProtocolOptions(maxConnections=2)
    reactor.listenTCP(port, factory)
    reactor.run()





from src.arqLogger import start_websocket


# Callable function when message is received by the server
def on_new_log(log):
   print("Text message received: {0}".format(log))


# Simple start sever with port and callable function
start_websocket(8001, on_new_log)
# ws://177.231.134.94:8001/


from enum import Enum
import queue
import uuid
from threading import Thread
import time


class Position(object):

    def __init__(self):
        self.size = 0


class Broker(object):

    def __init__(self):
        self.position = dict()

    def getposition(self, data, clone=False):
        return self.position[data.tradecontract]

    def update_position(self, contract, size):
        if contract not in self.position.keys():
            self.position[contract] = Position()
        self.position[contract].size += size

class Strategy(object):

    instruments = {}

    def __init__(self):
        self.set_instruments(Strategy.instruments)
        self.orders_new = queue.Queue()
        self.orders_replace = queue.Queue()
        self.orders_cancel = queue.Queue()
        self.orders_by_id = dict()
        self.broker = Broker()

    def get_order(self, index):
        idx = 0
        for order in self.orders_by_id.values():
            if idx == index:
                return order
            idx += 1

    def notify_submitted(self, order):
        time.sleep(0.5)
        order.status = Order.Submitted
        self.notify_order(order)
        
    def logs(self, msg):
        print("stratey log: "+msg)

    def set_instruments(self, instr):
        self.__instr__ = dict()
        for k, v in instr.items():
            self.__instr__[v.tradecontract] = k
            self.__setattr__(k, v)

    def buy(self, data, price, size, exectype):
        print("se envia una orden BUY para %s de %s@%s tipo %s." % (data.tradecontract,
                                                                    size, price,
                                                                    exectype))
        return self.create_order(data, price, size, exectype, Order.Buy)

    def sell(self, data, price, size, exectype):
        print("se envia una orden SELL para %s de %s@%s tipo %s." % (data.tradecontract,
                                                                     size, price,
                                                                     exectype))
        return self.create_order(data, price, size, exectype, Order.Sell)

    def replace(self, size, price, order):
        print("se envia a reemplazar orden %s id: %s - Pasamos de %s@%s a %s@%s." % (
            order.ordtype, order.m_orderId,
            order.size, order.price, size, price))
        order = OrderLimit(price, size, order.ordtype, data=order.data, ordType=order.type, replace_id=order.m_orderId)
        self.orders_by_id[order.m_orderId] = order
        self.orders_replace.put(order)
        return order

    def cancel(self, order):
        print("se envia a cancelar orden %s id: %s." % (order.ordtype, order.m_orderId))
        self.orders_cancel.put(self.orders_by_id[order.m_orderId])

    def create_order(self, data1, price, size, type, side):
        order = OrderLimit(price, size, side, data=data1, ordType=type)
        self.orders_new.put(order)
        self.orders_by_id[order.m_orderId] = order
        #Thread(target=self.notify_submitted, args=(order,)).start()
        return order

    def next(self):
        raise NotImplementedError("Must override next")

    def update_order(self, ord_id, state, size=0):
        time.sleep(0.1)
        order = self.orders_by_id[ord_id]
        order.status = state

        if order.status is not Order.Rejected and order.replace_id:
            del self.orders_by_id[order.replace_id]

        m = 1
        if order.ordtype == Order.Sell:
            m = -1
        if order.status is Order.Partial:
            self.broker.update_position(order.data.tradecontract, m*size)
            order.executed.exbits.append(Trade(order.price, size))
            order.remaining -= size
        elif order.status is Order.Completed:
            self.broker.update_position(order.data.tradecontract, m*size)
            order.executed.exbits.append(Trade(order.price, order.remaining))
            order.remaining -= 0.0

        if order.status in (Order.Canceled, Order.Completed, Order.Expired, Order.Rejected):
            del self.orders_by_id[order.m_orderId]

        self.notify_order(order)

    def update_md(self, instrument, bid_qty=0, bid_px=0, offer_px=0, offer_qty=0):
        time.sleep(0.1)
        if instrument in self.__instr__.keys():
            data = getattr(self, self.__instr__[instrument])
            data.update_bid_px(bid_px)
            data.update_bid_qty(bid_qty)
            data.update_offer_px(offer_px)
            data.update_offer_qty(offer_qty)
            self.next()
        else:
            raise Exception("Instrument not in the List of the Strategy")


class Settle(Enum):
    S_CI = 1
    S_24 = 2
    S_48 = 3
    S_72 = 4


class Order(Enum):
    Created = 0
    Submitted = 1
    Completed = 2
    Canceled = 3
    Rejected = 4
    Expired = 5
    Partial = 6
    Accepted = 7
    Replace = 8
    Limit = 9
    Market = 10
    Buy = 11
    Sell = 12


class OrderLimit(object):
    def __init__(self, px, qty, side, data=None, ordType=Order.Limit, id=None, replace_id=None):
        self.data = data
        self.m_orderId = id if id else uuid.uuid1().__str__()
        self.price = px
        self.size = qty
        self.ordtype = side
        self.m_action = "BUY" if side == 0 else "SELL"
        self.type = ordType
        self.status = Order.Submitted
        self.executed = Executed()
        self.remaining = qty
        self.replace_id = replace_id


class Executed:
    def __init__(self):
        self.exbits = []
        self.remsize = 0


class Trade:
    def __init__(self, px, qty):
        self.price = px
        self.size = qty


class MarketData(object):
    def __init__(self, name, price_size, min_size, tick_px, settlement, bid_px=0, bid_qty=0, offer_px=0, offer_qty=0):
        self.price_size = float(price_size)
        self.tradecontract = name
        self.tick_px = tick_px
        self.contractsize = min_size
        self.settlement = settlement.value
        self.offer_px = [0]
        self.bid_px = [0]
        self.offer_qty = [0]
        self.bid_qty = [0]
        self.offer_px[0] = float(offer_px)
        self.offer_qty[0] = float(offer_qty)
        self.bid_px[0] = float(bid_px)
        self.bid_qty[0] = float(bid_qty)

    def update_bid_px(self, bid_px):
        self.bid_px[0] = float(bid_px) if bid_px else None

    def update_bid_qty(self, bid_qty):
        self.bid_qty[0] = float(bid_qty) if bid_qty else None

    def update_offer_px(self, offer_px):
        self.offer_px[0] = float(offer_px) if offer_px else None

    def update_offer_qty(self, offer_qty):
        self.offer_qty[0] = float(offer_qty) if offer_qty else None

    def update_md(self, bid_px, bid_qty, offer_px, offer_qty):
        self.offer_px[0] = float(offer_px)
        self.offer_qty[0] = float(offer_qty)
        self.bid_px[0] = float(bid_px)
        self.bid_qty[0] = float(bid_qty)

    def get_tick(self):
        return self.tick_px


Instruments = {
    "AY24 - CI": MarketData("AY24 - CI", 100, 1, 0.5, Settle.S_CI),
    "AY24D - CI": MarketData("AY24D - CI", 100, 1, 0.01, Settle.S_CI),
    "AY24 - 24hs": MarketData("AY24 - 24hs", 100, 1, 0.5, Settle.S_24),
    "AY24D - 24hs": MarketData("AY24D - 24hs", 100, 1, 0.01, Settle.S_24),
    "AY24 - 48hs": MarketData("AY24 - 48hs", 100, 1, 0.5, Settle.S_48),
    "AY24D - 48hs": MarketData("AY24D - 48hs", 100, 1, 0.01, Settle.S_48),
    "PARY - CI": MarketData("PARY - CI", 100, 1000, 0.5, Settle.S_CI),
    "PARYD - CI": MarketData("PARYD - CI", 100, 1000, 0.01, Settle.S_CI),
    "PARY - 24hs": MarketData("PARY - 24hs", 100, 1000, 0.5, Settle.S_24),
    "PARYD - 24hs": MarketData("PARYD - 24hs", 100, 1000, 0.01, Settle.S_24),
    "PARY - 48hs": MarketData("PARY - 48hs", 100, 1000, 0.5, Settle.S_48),
    "PARYD - 48hs": MarketData("PARYD - 48hs", 100, 1000, 0.01, Settle.S_48),
    "PESOS - T+1": MarketData("PESOS - T+1", 1, 1, 0.01, Settle.S_24),
    "PESOS - T+2": MarketData("PESOS - T+2", 1, 1, 0.01, Settle.S_48),
    "GGAL - CI": MarketData("GGAL - CI", 1, 1, 0.05, Settle.S_CI),
    "GGAL - 48hs": MarketData("GGAL - 48hs", 1, 1, 0.05, Settle.S_48),
}



from enum import Enum


class Side:
    Buy = 0
    Sell = 1


class Order:
    Created = 0
    Submitted = 1
    Accepted = 2
    Partial = 3
    Completed = 4
    Canceled = 5
    Expired = 6
    Margin = 7
    Rejected = 8


class LogTypes(Enum):
    Event = "EV"
    Response = "RE"
    Own = "PR"


class EventTypes(Enum):
    Initialize = "INI"
    MarketData = "MD"
    ExecReport = "OR"
    Pause = "PA"
    Error = "ER"
    Internal = "INT"


class ResponseTypes(Enum):
    NewOrder = "NO"
    CancelOrder = "CO"
    ReplaceOrder = "RO"

